/***************************************************************************
 *   Copyright (C) 2005 by John Bennett                                    *
 *   n4xi@arrl.net                                                         *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//////////////////////////////////////////////////////////////////////////////
// makecallbk - merges FCC data from files 'EN.dat' and 'HD.dat' into table
// that contains only active licensees. If the Canadian database is present
// (AMATEUR.RPT), that data is also merged.
//
// Table created is named 'cbtemp' in the wxn database. Prior to running
// the program, a script named 'makefcc.sql' must be run as user postgres
// from the PostgreSQL interactive terminal (psql).
//

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef BOOL
    typedef int BOOL;
#endif

#ifndef TRUE
    #define TRUE 1
    #define FALSE 0
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libpq-fe.h>
#include <time.h>

// libwxn
#include <titlecase.h>
#include <morph.h>
#include <strxcpy.h>
#include <trim.h>

#define MAX_ERRORS 100

#define EN_FIELD_COUNT 24
#define HD_FIELD_COUNT 49

#define LINE_SIZE 512
#define UID_SIZE 16
#define CALL_SIZE 16
#define NAME_SIZE 224
#define CITY_SIZE 64
#define SP_SIZE 4
#define STATUS_SIZE 1

// local prototype
void Escape( char * pSource, int nMax );
BOOL record( char * psLine, int nMaxLen, int nFields, FILE * hIn, char * psCheck );
char * Parse( char * psSource, char * psField, int nFLen );
void Filter( char * psString );

static char g_szLine[LINE_SIZE+1];
static char g_szExec[LINE_SIZE+1];
static char g_szSource[LINE_SIZE+1];

int main(int argc, char *argv[])
{
    PGconn * pConn;
    PGresult * pResult;
    FILE * hFile;
    unsigned uRecords;
    unsigned uAlienRejects;
    int nErrorCount;
    int i;
    int nUid;
    char szUid[UID_SIZE+1];
    char szCall[CALL_SIZE+1];
    char szName[NAME_SIZE+1];
    char szCity[CITY_SIZE+1];
    char szSP[SP_SIZE+1];
    char szStatus[STATUS_SIZE+1];
    char szPath[256];
    char * pRec;
    BOOL bBackground;
    timespec ts;

    fprintf( stdout, "US-Canadian create/merge callbook utility for wxn. Version %s\n", VERSION );

    bBackground = FALSE;
    if( argc == 2 )
    {
        if( strcmp( argv[1], "--background" ) == 0 )
            bBackground = TRUE;
    }

    if( bBackground )
        fprintf( stdout, "Background processing enabled\n" );

    pConn = PQconnectdb( "dbname=wxn" );
    if( pConn == NULL )
    {
        fprintf( stderr, "Error! Unable to connect to database. Is PostgreSQL running?\n" );
        return -1;
    }
    if( PQstatus( pConn ) != CONNECTION_OK )
    {
        fprintf( stderr, "Error! Database connection status failed. Check PostgreSQL status.\n" );
        PQfinish( pConn );
        return -1;
    }

    //////////////////////////////////////////// EN table

    hFile = fopen( "EN.dat", "r" );
    if( hFile == NULL )
    {
        fprintf( stderr, "Error! File <%s> not found.\n", szPath );
        return -1;
    }
    else
        fprintf( stdout, "Processing 'EN.dat' (US data)\n" );
    uRecords = 0;
    nErrorCount = 0;

    //////////////////////// setup timespec struct for nanosleep
    ts.tv_sec = 0;
    ts.tv_nsec = 5000000; // 50 mS

    for(;;)     // loop until done
    {
        if( bBackground )
            nanosleep( &ts, NULL );

        if( !record( g_szLine, LINE_SIZE, EN_FIELD_COUNT, hFile, "EN" ) )
            break;      // EOF

        // initialize
        strcpy( g_szSource, g_szLine );
        Filter( g_szLine );
        pRec = g_szLine;
        nUid = 0;
        memset( szUid, 0, UID_SIZE );
        memset( szCall, 0, CALL_SIZE );
        memset( szName, 0, NAME_SIZE );
        memset( szCity, 0, CITY_SIZE );
        memset( szSP, 0, SP_SIZE );

        pRec = Parse( pRec, NULL, 0 );
        if( pRec == NULL )
            break;

        // uid
        pRec = Parse( pRec, szUid, UID_SIZE );
        if( pRec == NULL )
            break;
        nUid = atoi( szUid );

        // skip the next two
        for( i=0; i<2; i++ )
            pRec = Parse( pRec, NULL, 0 );
        if( pRec == NULL )
            break;

        // callsign
        pRec = Parse( pRec, szCall, CALL_SIZE );
        if( pRec == NULL )
            break;

        // skip the next two
        for( i=0; i<2; i++ )
            pRec = Parse( pRec, NULL, 0 );
        if( pRec == NULL )
            break;

        // entity name
        pRec = Parse( pRec, szName, NAME_SIZE );
        if( pRec == NULL )
            break;
        Escape( szName, NAME_SIZE );
        TitleCase( szName );

        // skip the next eight
        for( i=0; i<8; i++ )
            pRec = Parse( pRec, NULL, 0 );
        if( pRec == NULL )
            break;

        // city
        pRec = Parse( pRec, szCity, CITY_SIZE );
        if( pRec == NULL )
            break;
        Escape( szCity, CITY_SIZE );
        TitleCase( szCity );

        // state
        pRec = Parse( pRec, szSP, SP_SIZE );
        if( pRec == NULL )
            break;

        sprintf( g_szExec, "INSERT INTO encall( uid,callsign,name,city,state ) VALUES( %i,'%s','%s','%s','%s' )",
            nUid, szCall, szName, szCity, szSP );
        pResult = PQexec( pConn, g_szExec );
        if( PQresultStatus( pResult ) != PGRES_COMMAND_OK )
        {
            PQclear( pResult );
            fprintf( stderr, "Error! Unable to insert record into 'encall' table\n%s\n"
                "[%i] [%s] [%s] [%s] [%s]\n", g_szSource, nUid, szCall, szName, szCity, szSP );
            if( ++nErrorCount > MAX_ERRORS )
            {
                PQfinish( pConn );
                fclose( hFile );
                fprintf( stderr, "WHOA! Too many insertion errors!\n"
                    "Are you executing as user 'postgres' and did you run\n"
                    "'psql wxn -f makefcc.sql' first?\n" );
                return -1;
            }
            continue;
        }
        PQclear( pResult );
        uRecords++;
    }

    fprintf( stdout, "%u records added\n", uRecords );

    fclose( hFile );

    //////////////////////////////////////////// HD table

    hFile = fopen( "HD.dat", "r" );
    if( hFile == NULL )
    {
        fprintf( stderr, "Error! File <%s> not found.\n", szPath );
        PQfinish( pConn );
        return -1;
    }
    else
        fprintf( stdout, "Processing 'HD.dat' (US data)\n" );

    uRecords = 0;
    nErrorCount = 0;

    for(;;)     // loop until done
    {
        if( bBackground )
            nanosleep( &ts, NULL );

        if( !record( g_szLine, LINE_SIZE, HD_FIELD_COUNT, hFile, "HD" ) )
            break;      // EOF

        // initialize
        strcpy( g_szSource, g_szLine );
        Filter( g_szLine );
        pRec = g_szLine;
        nUid = 0;
        memset( szUid, 0, UID_SIZE );
        memset( szCall, 0, CALL_SIZE );
        memset( szStatus, 0, STATUS_SIZE );

        pRec = Parse( pRec, NULL, 0 );
        if( pRec == NULL )
            break;

        // uid
        pRec = Parse( pRec, szUid, UID_SIZE );
        if( pRec == NULL )
            break;
        nUid = atoi( szUid );

        // skip the next two
        for( i=0; i<2; i++ )
            pRec = Parse( pRec, NULL, 0 );
        if( pRec == NULL )
            break;

        // callsign
        pRec = Parse( pRec, szCall, CALL_SIZE );
        if( pRec == NULL )
            break;

        // status
        pRec = Parse( pRec, szStatus, STATUS_SIZE );
        if( pRec == NULL )
            break;

        // skip the remainder
        for( i=0; i<43; i++ )
            pRec = Parse( pRec, NULL, 0 );
        if( pRec == NULL )
            break;

        sprintf( g_szExec, "INSERT INTO hdcall( uid,callsign,status ) VALUES( %i,'%s','%c' )",
            nUid, szCall, *szStatus );
        pResult = PQexec( pConn, g_szExec );
        if( PQresultStatus( pResult ) != PGRES_COMMAND_OK )
        {
            PQclear( pResult );
            fprintf( stderr, "Error! Unable to insert record into 'hdcall' table\n%s\n"
                "[%i] [%s] [%c]\n", g_szSource, nUid, szCall, *szStatus );
            if( ++nErrorCount > MAX_ERRORS )
            {
                PQfinish( pConn );
                fclose( hFile );
                fprintf( stderr, "Too many insertion errors. Exiting.\n" );
                return -1;
            }
            continue;
        }
        PQclear( pResult );
        uRecords++;
    }

    fprintf( stdout, "%u records added\n", uRecords );

    fclose( hFile );

    //////////////////////////////////////////// create 'cbtemp' table

    fprintf( stdout, "Merging EN and HD data into new table 'cbtemp'\n" );

    sprintf( g_szExec,
        "CREATE TABLE cbtemp AS SELECT e.callsign,e.name,e.city,e.state FROM encall AS e, hdcall AS h WHERE e.uid = h.uid AND h.status='A'" );
    pResult = PQexec( pConn, g_szExec );
    if( PQresultStatus( pResult ) != PGRES_COMMAND_OK )
    {
        fprintf( stderr, "Error! Unable to create 'cbtemp' table.\n" );
        PQclear( pResult );
        PQfinish( pConn );
        return -1;
    }
    PQclear( pResult );

    sprintf( g_szExec, "SELECT count(*) from cbtemp" );
    pResult = PQexec( pConn, g_szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        fprintf( stdout, "Merging US data created %s records in cbtemp table\n", PQgetvalue( pResult, 0, 0 ) );
    else
        fprintf( stderr, "  ERROR! Unable to get record count on 'cbtemp' table.\n" );
    PQclear( pResult );

    //////////////////////////////////////////// add canadian database

    hFile = fopen( "AMATEUR.RPT", "r" );
    if( hFile == NULL )
    {
        fprintf( stderr, "Error! File <%s> not found.\n", szPath );
        return -1;
    }
    else
        fprintf( stdout, "Processing 'AMATEUR.RPT' (Canadian data)\n" );

    uRecords = 0;
    uAlienRejects = 0;
    nErrorCount = 0;

    // skip first four lines
    for( i=0; i<4; i++ )
        fgets( g_szLine, LINE_SIZE, hFile );

    for(;;)     // loop until done
    {
        if( bBackground )
            nanosleep( &ts, NULL );

        if( fgets( g_szLine, LINE_SIZE, hFile ) == NULL )
            break;      // EOF
        if( memcmp( g_szLine, "EOF", 3 ) == 0 )
            break;      // also EOF

        // strip CR/LFs
        pRec = strchr( g_szLine, '\r' );
        if( pRec )
            *pRec = 0;
        pRec = strchr( g_szLine, '\n' );
        if( pRec )
            *pRec = 0;

        strcpy( g_szSource, g_szLine );
        Filter( g_szLine );

        memset( szCall, 0, CALL_SIZE );
        memset( szName, 0, NAME_SIZE );
        memset( szCity, 0, CITY_SIZE );
        memset( szSP, 0, SP_SIZE );

        // callsign
        strxcpy( szCall, g_szLine, 6 );
        TrimSpaces( szCall );

        // name
        strxcpy( szName, g_szLine+7, 35 );
        TrimSpaces( szName );
        strxcpy( g_szExec, g_szLine+43, 35 );
        TrimSpaces( g_szExec );
        pRec = strchr( g_szExec, '(' );
        if( pRec )
            *pRec = 0;      // get rid of trailing '(+)' if it exists
        strcat( szName, " " );
        strcat( szName, g_szExec );
        Escape( szName, NAME_SIZE );

        // city
        strxcpy( szCity, g_szLine+150, 35 );
        TrimSpaces( szCity );
        TitleCase( szCity );
        Escape( szCity, CITY_SIZE );

        // province
        strxcpy( szSP, g_szLine+186, 2 );
        TrimSpaces( szSP );
        if( memcmp( szSP, "ZZ", 2 ) == 0 )
        {
            uAlienRejects++;
            continue;       // don't add non-canadian licensees
        }

        sprintf( g_szExec, "INSERT INTO cbtemp( callsign,name,city,state ) VALUES( '%s','%s','%s','%s' )",
            szCall, szName, szCity, szSP );
        pResult = PQexec( pConn, g_szExec );
        if( PQresultStatus( pResult ) != PGRES_COMMAND_OK )
        {
            PQclear( pResult );
            fprintf( stderr, "    ERROR! Unable to insert record into 'cbtemp' table\n%s\n"
                "Call:[%s] Name:[%s] City:[%s] State/Prov:[%s]\n", g_szSource, szCall, szName, szCity, szSP );
            if( ++nErrorCount > MAX_ERRORS )
            {
                PQfinish( pConn );
                fclose( hFile );
                fprintf( stderr, "Too many insertion errors. Exiting.\n" );
                return -1;
            }
            continue;
        }
        PQclear( pResult );
        uRecords++;
    }

    fprintf( stdout, "%u records added (%u ignored w/non-Canadian address)\n", uRecords, uAlienRejects );
    fclose( hFile );

    sprintf( g_szExec, "SELECT count(*) from cbtemp" );
    pResult = PQexec( pConn, g_szExec );
    if( (PQresultStatus( pResult ) == PGRES_TUPLES_OK) && PQntuples( pResult ) )
        fprintf( stdout, "Merging canadian data resulted in %s records in cbtemp table\n", PQgetvalue( pResult, 0, 0 ) );
    else
        fprintf( stderr, "Error! Unable to get record count on 'cbtemp' table.\n" );
    PQclear( pResult );

    PQfinish( pConn );

    return EXIT_SUCCESS;
}

//////////////////////////////////////////////////////////////////////////////
// Escape - put proper escape sequences in text stream for SQL
//

void Escape( char * pSource, int nMax )
{
    char szString[256];
    char * pS;
    char * pD;

    // escape the single quote '
    if( strchr( pSource, '\'' ) )
    {
        pS = pSource;
        memset( szString, 0, 256 );
        pD = szString;

        while ( *pS )
        {
            if ( *pS == '\'' )
                *pD++ = *pS;    // escape the ' SQL style
            *pD++ = *pS++;
        }
        strxcpy( pSource, szString, nMax );
    }

    // remove any trailing escape characters
    if( strchr( pSource, '\\' ) )
    {
        pS = pSource;
        memset( szString, 0, 256 );
        pD = szString;

        while ( *pS )
        {
            if ( *pS == '\\' && *(pS +1) == 0 )
                break;
            else
                *pD++ = *pS++;
        }
        strxcpy( pSource, szString, nMax );
    }

}

/////////////////////////////////////////////////////////////////////////////
// record - reads one complete record from input, ignoring CRs and LFs.
// terminates only when number of fields specified are read or EOF is
// encountered. Buffer is zeroed completely before starting. Fields in
// a record are delimited by '|'. 'psCheck' contains record identifier
// which appears in the first field of the record. Returns TRUE success,
// FALSE on failure.
//

BOOL record( char * psLine, int nMaxLen, int nFields, FILE * hIn, char * psCheck )
{
    char * pS;
    char cCh;
    int nLen = 0;
    int nFieldCount = 0;

    memset( psLine, 0, nMaxLen );
    pS = psLine;
    for(;;)     // FOREVER
    {
        if( fread( &cCh, 1, 1, hIn ) == 0 )
            break;      // EOF
        if( cCh == '\r' || cCh == '\n' )
            continue;
        *pS = cCh;
        if( *pS == '|' )
            nFieldCount++;
        pS++;
        if( nFieldCount == nFields )
        {
            if ( memcmp( psLine, psCheck, 2 ) )
            {
                fprintf( stderr, "Record header failed sanity check\n" );
                break;
            }
            *pS = 0;
            return TRUE;
        }
        nLen++;
        if( nLen >= nMaxLen-1 )
        {
            fprintf( stderr, "Input buffer overflow\n" );
            break;
        }
    }

    *pS = 0;
    return FALSE;
}

/////////////////////////////////////////////////////////////////////////////
// Parse - parses field from record. Delimiter is '|'. Delimiters are always
// trailing. If 'psField' is NULL, just finds beginning of next field or
// NULL if there is no next field or error.
//

char * Parse( char * psSource, char * psField, int nFLen )
{
    char * pS;
    char * pF;
    int nLen;

    if( psSource == NULL )
        return NULL;

    pS = psSource;
    if( psField )
        pF = psField;
    else
        pF = NULL;

    nLen = 0;

    while( *pS )
    {
        if( nLen > nFLen )
        {
            *pF = 0;
            return NULL;
        }
        if( *pS == '|' )
            break;
        if( pF )
        {
            *pF = *pS;
            pF++;
            nLen++;
        }
        pS++;
    }

    if( pF )
        *pF = 0;
    pS++;   // point past '|'

    return pS;
}


//////////////////////////////////////////////////////////////////////////////
// Filter - replace characters outside the range of normal ASCII with
// appropriate substitute or '_' to avoid postgresql complaints.
//

void Filter( char * psString )
{
    char * pS;

    pS = psString;
    while( *pS )
    {
        if( *pS > 0x7e || *pS < 0x20 )
        {
            switch( *pS )
            {
                case 0xc0:
                case 0xc2:
                    *pS = 'A';
                    break;
                case 0xe6:
                    *pS = 'a';
                    break;
                case 0xc7:
                    *pS = 'C';
                    break;
                case 0xe7:
                    *pS = 'c';
                    break;
                case 0xc9:
                case 0xc8:
                case 0xca:
                case 0xcb:
                    *pS = 'E';
                    break;
                case 0xe9:
                case 0xeb:
                case 0xe8:
                case 0xea:
                    *pS = 'e';
                    break;
                case 0xee:
                case 0xce:
                case 0xef:
                    *pS = 'i';
                    break;
                case 0xf1:
                    *pS = 'N';
                    break;
                case 0xd8:
                case 0xd4:
                    *pS = 'O';
                    break;
                case 0xf3:
                case 0xf4:
                    *pS = 'o';
                    break;
                case 0xdb:
                    *pS = 'U';
                    break;
                case 0xfa:
                case 0xfc:
                    *pS = 'u';
                    break;

                default:
                    *pS = '_';
                    break;
            }
        }
        pS++;
    }
}
