//---------------------------------------------------------------------------
// Copyright (C) 2000 Dallas Semiconductor Corporation, All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
// OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of Dallas Semiconductor
// shall not be used except as stated in the Dallas Semiconductor
// Branding Policy.
//--------------------------------------------------------------------------
//
//  atod26.c - Reads the voltage on the 1-Wire of the DS2438.
//  version 1.00
//

// libwxn
#include <logmsg.h>

#include "ownet.h"
#include "atod26.h"

#ifndef BOOL
    typedef int BOOL;
#endif

/**
 * Sets the DS2438 to read Vad or Vdd
 *
 * vdd      flag indicating weather to read the Vdd or the Vad
 * pSN      the serial number for the part to be read
 *
 * @return 'true' if the read was complete
 */

BOOL SetupAtoD( int vdd, uchar *pSN )
{
    uchar data_block[50];
    uchar test;
    int send_cnt;
    int i;
    ushort lastcrc8;

    send_cnt=0;

    owSerialNum( pSN, FALSE );
    // Recall the Status/Configuration page
    // Recall command
    data_block[send_cnt++] = 0xB8;

    // Page to Recall
    data_block[send_cnt++] = 0x00;

    if( !owBlock(FALSE, data_block, send_cnt) )
        return FALSE;

    if( owAccess() )
    {
        send_cnt = 0;

        // Read the Status/Configuration byte
        // Read scratchpad command
        data_block[send_cnt++] = 0xBE;

        // Page for the Status/Configuration byte
        data_block[send_cnt++] = 0x00;

        for( i=0; i<9; i++ )
            data_block[send_cnt++] = 0xFF;

        if( owBlock(FALSE,data_block,send_cnt) )
        {
            setcrc8( 0 );
            lastcrc8 = 0xff;    // assume failure

            for( i=2; i<send_cnt; i++ )
                lastcrc8 = docrc8( data_block[i] );

            if( lastcrc8 != 0x00 )
            {
                g_log.PostMsg( HOSTMSG_CRC_FAIL, "SetupAtoD()" );
                return FALSE;
            }
        }
        else
            return FALSE;

        test = data_block[2] & 0x08;
        if(((test == 0x08) && vdd) || ((test == 0x00) && !(vdd)))
            return TRUE;
    }

    if( owAccess() )
    {
        send_cnt = 0;
        // Write the Status/Configuration byte
        // Write scratchpad command
        data_block[send_cnt++] = 0x4E;

        // Write page
        data_block[send_cnt++] = 0x00;

        if( vdd )
            data_block[send_cnt++] = data_block[2] | 0x08;
        else
            data_block[send_cnt++] = data_block[2] & 0xF7;

        for( i=0; i<7; i++ )
            data_block[send_cnt++] = data_block[i+4];

        if(owBlock(FALSE,data_block,send_cnt))
        {
            send_cnt = 0;

            if( owAccess() )
            {
                // Copy the Status/Configuration byte
                // Copy scratchpad command
                data_block[send_cnt++] = 0x48;

                // Copy page
                data_block[send_cnt++] = 0x00;

                if( owBlock(FALSE,data_block,send_cnt) )
                {
                    while( !owReadByte() );
                    return TRUE;
                }
            }
        }
    }

   return FALSE;

}


/**
 * Read the Vdd or the Vad from the DS2438
 *
 * vdd      flag indicating weather to read the Vdd or the Vad
 * pSN      the serial number for the part to be read
 *
 * @return the integer value of Vad or Vdd
 */

int ReadAtoD( int vdd, uchar * pSN )
{
    uchar data_block[50];
    int send_cnt;
    int i;
    ushort lastcrc8;
    int nVal;
    int ret;
    BOOL bDone;

    bDone = TRUE;
    ret = 0;

    do
    {
        if( SetupAtoD(vdd,pSN) )
        {

            if( owAccess() )
            {
                if( !owWriteByte(0xB4) )
                    return ret;
                while( !owReadByte() );
            }

            if( owAccess() )
            {
                send_cnt = 0;

                // Recall the Status/Configuration page
                // Recall command
                data_block[send_cnt++] = 0xB8;

                // Page to Recall
                data_block[send_cnt++] = 0x00;

                if( !owBlock(FALSE, data_block, send_cnt) )
                    return ret;
            }

            if( owAccess() )
            {

                send_cnt = 0;

                // Read the Status/Configuration byte
                // Read scratchpad command
                data_block[send_cnt++] = 0xBE;

                // Page for the Status/Configuration byte
                data_block[send_cnt++] = 0x00;

                for( i=0; i<9; i++ )
                    data_block[send_cnt++] = 0xFF;

                if( owBlock(FALSE, data_block, send_cnt) )
                {
                    setcrc8( 0 );
                    lastcrc8 = 0xff;    // assume failure

                    for( i=2; i<send_cnt; i++ )
                        lastcrc8 = docrc8( data_block[i] );

                    if( lastcrc8 != 0x00 )
                    {
                        g_log.PostMsg( HOSTMSG_CRC_FAIL, "ReadAtoD()" );
                        return ret;
                    }
                }
                else
                    return ret;

                if( (!vdd) && ((data_block[2] & 0x08) == 0x08) )
                    bDone = FALSE;
                else
                    bDone = TRUE;

                nVal = (data_block[6] << 8) | data_block[5];
                return nVal;

            }
        }

    } while( !bDone );

    return ret;

}


/**
 * Reads the temperature from the DS2438.
 *
 * pSN      serial number for the part to be read
 *
 * @return the temperature in returned in C.
 */

BOOL DS2438Temperature( uchar * pSN, int * pVal )
{
    short int nVal;
    short int nValMask;
    uchar data_block[50];
    int send_cnt;
    int i;
    ushort lastcrc8;

    owSerialNum( pSN,FALSE );

    if( owAccess() )
    // Convert Temperature command
    owWriteByte( 0x44 );

    msDelay( 20 );

    if( owAccess() )
    {
        send_cnt = 0;

        // Recall the Status/Configuration page
        // Recall command
        data_block[send_cnt++] = 0xB8;

        // Page to Recall
        data_block[send_cnt++] = 0x00;

        if( !owBlock(FALSE, data_block, send_cnt) )
            return FALSE;   // owBlock will write error to log
    }

    if( owAccess() )
    {

        send_cnt = 0;

        // Read the Status/Configuration byte
        // Read scratchpad command
        data_block[send_cnt++] = 0xBE;

        // Page for the Status/Configuration byte
        data_block[send_cnt++] = 0x00;

        for( i=0; i<9; i++ )
            data_block[send_cnt++] = 0xFF;

        if( owBlock(FALSE,data_block,send_cnt) )
        {
            setcrc8(0);
            lastcrc8 = 0xff;    // assume failuire

            for( i=2; i<send_cnt; i++ )
                lastcrc8 = docrc8( data_block[i] );
            if( lastcrc8 != 0x00 )
            {
                g_log.PostMsg( HOSTMSG_CRC_FAIL, "DS2438Temperature()" );
                return FALSE;
            }
        }
        else
            return FALSE;   // owBlock will write error to log

        nVal = (short) (data_block[4] << 8) | (short) data_block[3];
        if( nVal < 0 )
            nValMask = -1;
        else
            nValMask = 0;
        nVal = (nVal >> 3) & nValMask;

        if( (abs(nVal) % 32) >= 16 )
        {
            if( nVal < 0 )
                *pVal = (int) ( (nVal/32) - 1 );
            else
                *pVal = (int) ( (nVal/32) + 1 );
        }
        else
            *pVal = (int) ( nVal/32 );

        return TRUE;

    }   // owAccess will write error to log

    return FALSE;

}

