//---------------------------------------------------------------------------
// Copyright (C) 2000 Dallas Semiconductor Corporation, All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
// MERCHANTABILITY,  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
// IN NO EVENT SHALL DALLAS SEMICONDUCTOR BE LIABLE FOR ANY CLAIM, DAMAGES
// OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
// OTHER DEALINGS IN THE SOFTWARE.
//
// Except as contained in this notice, the name of Dallas Semiconductor
// shall not be used except as stated in the Dallas Semiconductor
// Branding Policy.
//---------------------------------------------------------------------------
//
//  ds2480ut.c - DS2480B utility functions.
//
//  Version: 2.01
//
//  History: 1.00 -> 1.01  Default PDSRC changed from 0.83 to 1.37V/us
//                         in DS2480Detect. Changed to use msDelay instead
//                         of Delay.
//           1.01 -> 1.02  Changed global declarations from 'uchar' to 'int'.
//                         Changed DSO/WORT from 7 to 10us in DS2480Detect.
//           1.02 -> 1.03  Removed caps in #includes for Linux capatibility
//           1.03 -> 2.00  Changed 'MLan' to 'ow'. Added support for
//                         multiple ports.  Changed W1LT to 8us.
//           2.00 -> 2.01 Added error handling. Added circular-include check.
//           2.01 -> 2.10 Added raw memory error handling and int
//           2.10 -> 3.00 Added memory bank functionality
//                        Added file I/O operations
//

#define _DS2480UT_CPP_

#include "ownet.h"
#include "ds2480.h"

// global DS2480B state
int g_nLevel; // current DS2480B 1-Wire Net level
int g_nBaud;  // current DS2480B baud rate
int g_nMode;  // current DS2480B command or data mode state
int g_nSpeed; // current DS2480B 1-Wire Net communication speed


//---------------------------------------------------------------------------
// Attempt to resyc and detect a DS2480B
//
// Returns:  TRUE  - DS2480B detected successfully
//           FALSE - Could not detect DS2480B
//

BOOL DS2480Detect( void )
{
   uchar sendpacket[10];
   uchar readbuffer[10];
   uchar sendlen=0;

   // reset modes
   g_nMode = MODSEL_COMMAND;
   g_nBaud = PARMSET_9600;
   g_nSpeed = SPEEDSEL_FLEX;

   // set the baud rate to 9600
   SetBaudCOM( (uchar) g_nBaud );

   // send a break to reset the DS2480
   BreakCOM();

   // delay to let line settle
   msDelay(2);

   // flush the buffers
   FlushCOM();

   // send the timing byte
   sendpacket[0] = 0xC1;
   if( !WriteCOM(1,sendpacket) )
   {
      OWERROR( OWERROR_WRITECOM_FAILED );
      return FALSE;
   }

   // delay to let line settle
   msDelay(4);

   // set the FLEX configuration parameters
   // default PDSRC = 1.37Vus
   sendpacket[sendlen++] = CMD_CONFIG | PARMSEL_SLEW | PARMSET_Slew1p37Vus;
   // default W1LT = 10us
   sendpacket[sendlen++] = CMD_CONFIG | PARMSEL_WRITE1LOW | PARMSET_Write10us;
   // default DSO/WORT = 8us
   sendpacket[sendlen++] = CMD_CONFIG | PARMSEL_SAMPLEOFFSET | PARMSET_SampOff8us;

   // construct the command to read the baud rate (to test command block)
   sendpacket[sendlen++] = CMD_CONFIG | PARMSEL_PARMREAD | (PARMSEL_BAUDRATE >> 3);

   // also do 1 bit operation (to test 1-Wire block)
   sendpacket[sendlen++] = CMD_COMM | FUNCTSEL_BIT | g_nBaud | BITPOL_ONE;

   // flush the buffers
   FlushCOM();

   // send the packet
   if (WriteCOM(sendlen,sendpacket) )
   {
      // read back the response
      if (ReadCOM(5,readbuffer) == 5 )
      {
         // look at the baud rate and bit operation
         // to see if the response makes sense
         if (((readbuffer[3] & 0xF1) == 0x00) &&
             ((readbuffer[3] & 0x0E) == g_nBaud) &&
             ((readbuffer[4] & 0xF0) == 0x90) &&
             ((readbuffer[4] & 0x0C) == g_nBaud))
            return TRUE;
         else
            OWERROR( OWERROR_DS2480_BAD_RESPONSE );
      }
      else
         OWERROR( OWERROR_READCOM_FAILED );
   }
   else
      OWERROR( OWERROR_WRITECOM_FAILED );

   return FALSE;
}

