/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <signal.h>
#include <string.h>
#include <netax25/daemon.h>
#include <syslog.h>

// libwxn
#include <logmsg.h>
#include <cconfig.h>
#include <tz.h>
#include <sun.h>

#include "umx.h"
#include "id5.h"
#include "dav2.h"
#include "davp.h"
#include "emp.h"
#include "precip.h"
#include "hconfig.h"
#include "wxport.h"
#include "wxdata.h"
#include "archive.h"
#include "qthinfo.h"
#include "onewire.h"
#include "wxpic.h"
#include "alarm.h"



static void signal_handler( int nSig );
static void ExitFail( int nMsg, char * sInfo );


// global data

BOOL g_bInitRates = FALSE;
BOOL g_bVerbose = FALSE;
static BOOL g_bStop = FALSE;

logmsg g_log;


static void signal_handler( int nSig )
{
    switch ( nSig )
    {
        case SIGHUP:
            g_log.PostMsg( HOSTMSG_READ_ALARM_CONFIG, NULL );
            ReadAlarmConfig();
            break;

        case SIGTERM:
        case SIGQUIT:
        case SIGKILL:
            g_bStop = TRUE;
            break;

        default:
            break;
    }
}


static void ExitFail( int nMsg, char * sInfo )
{
    char szTerminate[] = "Terminating - check app log";

    g_log.PostMsg( nMsg, sInfo );
    syslog( LOG_ERR, szTerminate );
    fprintf( stderr, "%s\n", szTerminate );
    exit( 1 );
}


int main( int argc, char *argv[] )
{
    BOOL bReadInProgress;
    BOOL bResetDataDone;
    BOOL bHourlyDone;
    BOOL bStartDaemon;

    bStartDaemon = TRUE;

    if( argc > 1 )
    {
        if( strcmp( argv[1], "--debug" ) == 0 )
        {
            g_bVerbose = TRUE;
            bStartDaemon = FALSE;
        }
        else if( strcmp(argv[1], "--verbose") == 0 )
            g_bVerbose = TRUE;
        else if( strcmp(argv[1], "--build") == 0 )
        {
            fprintf( stdout, "wxnhostd: ver %s (built %s @ %s)\n", VERSION, __DATE__, __TIME__ );
            exit( 0 );
        }
        else if( strcmp(argv[1], "--rates") == 0 )
            g_bInitRates = TRUE;
    }

    g_log.init( WXN_LOGDIR "wxnhostd.log", g_bVerbose );
    if( g_bVerbose )
        fprintf( stdout, "wxnhostd log file: %s\n", g_log.dir() );

    if( bStartDaemon )
    {
        if( !daemon_start(TRUE) )
            ExitFail( MSG_DAEMON_START_FAIL, NULL );
    }

    // set signal handlers
    signal(SIGTERM, signal_handler);
    signal(SIGHUP,  signal_handler);
    signal(SIGQUIT, signal_handler);
    signal(SIGPIPE, SIG_IGN);

    // print signon message
    if( g_bVerbose )
    {
        fprintf( stdout, "wxn host weather station daemon for Linux, ver %s (built %s @ %s)\n",
            VERSION, __DATE__, __TIME__ );
        fprintf( stdout, "Process ID: %i\n", (int) getpid() );
    }

    g_log.PostMsg( MSG_INIT_BEGIN, NULL );
    syslog( LOG_INFO, "started" );

    //////////////////////// read config files

    if( !ReadComConfig() )
        ExitFail( MSG_STARTUP_FAIL, NULL );

    if( !ReadHostConfig() )
        ExitFail( MSG_STARTUP_FAIL, NULL );

    //////////////////////// read/create local record as required

    if( !SetLocalInfo() )
        ExitFail( HOSTMSG_LOAD_CREATE_LOCAL_RECORD, NULL );

    //////////////////////// setup time variables

    g_Time.Init();
    g_Time.Set();

    //////////////////////// open wx stn COM port

    if( !OpenWxStn( g_HostConfig.nBaud, g_HostConfig.szWxStnPort, g_HostConfig.nTimeout ) )
        ExitFail( MSG_STARTUP_FAIL, NULL );

    //////////////////////// if required, open and configure one-wire support

    if( g_HostConfig.bEnableOneWire )
    {
        if( !OpenOneWire( g_HostConfig.szOneWirePort ) )
        {
            g_HostConfig.bEnableOneWire = FALSE;
            g_HostConfig.bOwEmp = FALSE;
            g_HostConfig.bOwRad = FALSE;
            g_log.PostMsg( HOSTMSG_OPEN_OW_PORT_FAIL, NULL );
        }
        else if( !IdOneWireDevices() )
            CloseOneWire();
    }

    //////////////////////// Initialize system

    if( !InitWxData( g_bVerbose ) )
    {
        CloseWxStn();
        CloseOneWire();
        ExitFail( MSG_INIT_FAIL, "data" );
    }

    ReadAlarmConfig();  // get current alarm trip points

    //////////////////////// Peet Weather Picture

    if( g_HostConfig.bEnableWxPic )
        OpenWxPic( g_HostConfig.szWxPicPort, g_bVerbose );

    //////////////////////////////////////////////////////////////////////////
    // initialize for main loop

    bReadInProgress = FALSE;
    bResetDataDone = FALSE;
    bHourlyDone = FALSE;
    ResetAlarms();

    g_log.PostMsg( MSG_STARTUP_OK, NULL );

    //////////////////////////////////////////////////////////////////////////
    // main loop
    //////////////////////////////////////////////////////////////////////////

    FOREVER     // for(;;)
    {
        if( g_bStop )
            break;

        // get current time
        g_Time.Set();

        if( g_Time.NewDay() )
        {
            if( !bResetDataDone )           // so we only do this once
            {
                bResetDataDone = TRUE;
                WriteDaily();
                SetWxRecords();
                if( g_Time.NewMonth() )
                {
                    WriteMonthly();
                    ResetMonthPrecip();
                }
                if( g_Time.NewYear() )
                    ResetAnnualPrecip();
                WxClear();                  // clear daily info
                WxStnClear();               // ditto for wx stn, if required
                CalculateRiseSet( &g_Local.Info, &g_Local.Data );
                SaveLocalSunData();
            }
        }
        else
            bResetDataDone = FALSE;

        // read wx station every HOST_INTERVAL minutes
        if( g_Time.Minute() % HOST_INTERVAL == 0 )
        {
            if( !bReadInProgress )
            {
                bReadInProgress = TRUE;
                if( GetCurrentWeatherData( TRUE ) )
                {
                    SaveLocalData();
                    DataAlarm();
                }
                else
                {
                    g_log.PostMsg( HOSTMSG_WXSTN_READ_FAIL, NULL );
                }
            }
        }
        else
            bReadInProgress = FALSE;

        // make hourly summary
        if( g_Time.Minute() == 0 )
        {
            if( !bHourlyDone )  // so we only do this once on the hour
            {
                bHourlyDone=TRUE;
                SetPrecipLastHour();
                WriteHourly();
                ResetAlarms();
                ReadAlarmConfig();  // re-read trip points only on the hour
                //// error logging - if commented out, not used
                switch( g_HostConfig.nStnType )
                {
                    //case ULT2000:
                    //case ULTII:
                    //    break;

                    //case ID5001:
                    //    break;

                    //case DAVIS2:
                    //    break;

                    case DAVISVP:
                        DavPLog();
                        break;

                    default:
                        break;
                }
            }
        }
        else
            bHourlyDone=FALSE;

        sleep( 1 );
    }

    //////////////////////////////////////////////////////////////////////////
    // exit program

    switch( g_HostConfig.nStnType )
    {
        case ULT2000:
        case ULTII:
            UmxStopThread( g_bVerbose );
            break;

        case ID5001:
            Id5StopThread( g_bVerbose );
            break;

        case DAVIS2:
            Dav2StopThread( g_bVerbose );
            break;

        case DAVISVP:
            DavPStopThread( g_bVerbose );
            break;

        default:
            break;
    }

    CloseWxStn();
    CloseOneWire();
    CheckAlarmThread();
    CloseWxPic( g_bVerbose );

    g_log.PostMsg( MSG_DAEMON_STOP, NULL );
    syslog( LOG_INFO, "stopped" );

    return EXIT_SUCCESS;
}

