/***************************************************************************
 *   Copyright (C) 2005 by John Bennett   *
 *   n4xi@arrl.net   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

//
// wxpic.cpp  -  manages serial port stream to drive Peet Brothers "Weather Picture"
//

#define _WXPIC_CPP_

#include <stdio.h>
#include <stdlib.h>
#include <termios.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <pthread.h>

// libwxn
#include <logmsg.h>
#include <tz.h>
#include <makeuc.h>

#include "wxpic.h"
#include "wxdata.h"
#include "hconfig.h"


// local prototype
int MphToKph( int nMph );


static int g_hWxPicPort = 0;
static struct termios g_OldWxPicIos;
static struct termios g_NewWxPicIos;

// thread related variables/functions
void * WritePic( void * pParam );
static BOOL g_bStopPicThread = FALSE;
static BOOL g_bPicThreadRunning = FALSE;
static pthread_t g_hPicThread;


//////////////////////////////////////////////////////////////////////////////
// OpenWxPic - open Weather Picture serial port and start thread
//

BOOL OpenWxPic( char * szPort, BOOL bVerbose )
{
    char szDev[32];
    int nStatus;

    sprintf( szDev, "/dev/%s", szPort );
    g_hWxPicPort = open( szDev, O_RDWR | O_NOCTTY | O_NONBLOCK );
    if( g_hWxPicPort < 0 )
    {
        g_hWxPicPort = 0;
        g_log.PostMsg( HOSTMSG_WXPIC_PORT_FAIL, szPort );
        g_HostConfig.bEnableWxPic = FALSE;
        return FALSE;
    }

    // save old termios
    tcgetattr( g_hWxPicPort, &g_OldWxPicIos );

    bzero( &g_NewWxPicIos, sizeof( g_NewWxPicIos ) );
    g_NewWxPicIos.c_cflag = B2400 | CLOCAL | CREAD | CS8;
    g_NewWxPicIos.c_iflag = IGNPAR | IGNBRK;

    tcflush( g_hWxPicPort, TCIFLUSH );
    tcsetattr( g_hWxPicPort, TCSANOW, &g_NewWxPicIos );

    ////////////// start the worker thread

    if( bVerbose )
        fprintf( stdout, "Starting WxPic worker thread..." );

    nStatus = pthread_create( &g_hPicThread, NULL, &WritePic, NULL );
    if( nStatus == 0 )
    {
        if( bVerbose )
            fprintf( stdout, "done.\n" );
        return TRUE;
    }

    fprintf( stderr, "Failed to start WxPic worker thread.\n" );
    return FALSE;
}


//////////////////////////////////////////////////////////////////////////////
// CloseWxPic - stop thread and close Weather Picture serial port
//

void CloseWxPic( BOOL bVerbose )
{
    if( g_bPicThreadRunning )
    {
        g_bStopPicThread = TRUE;
        pthread_join( g_hPicThread, NULL );
        g_bStopPicThread = FALSE;

        if( bVerbose )
            fprintf( stdout, "Weather Picture thread halted\n" );
    }

    if( g_hWxPicPort )
    {
        tcsetattr( g_hWxPicPort, TCSANOW, &g_OldWxPicIos );
        close( g_hWxPicPort );
        g_hWxPicPort = 0;
    }

}


//////////////////////////////////////////////////////////////////////////////
// WritePic - writes data to Weather Picture
//

void * WritePic( void * pParam )
{
    char szString[128];
    int nDir;
    int nSpd;
    int nGust;
    void * pBlackHole;

    pBlackHole = pParam;

    g_bPicThreadRunning = TRUE;

    //FOREVER
    for(;;)
    {

        sleep( 5 );

        nDir = (g_Local.Data.wac.dir * 255)/360;
        nSpd = MphToKph( g_Local.Data.wac.spd );
        nGust = MphToKph( g_Local.Data.wgc.spd );

        sprintf( szString, "!!%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x%04x\r\n",
            nSpd, nDir, g_Local.Data.tcr*10, g_Local.Data.pta, g_Local.Data.bcr,
            g_Local.Data.ist*10, g_Local.Data.hcr*10, g_Local.Data.ish*10,
            g_Time.YearDay(), g_Time.DayMinute(), g_Local.Data.pca, nGust );
        makeuc( szString );

        write( g_hWxPicPort, szString, strlen( szString ) );

        if( g_bStopPicThread )
        {
            g_bPicThreadRunning = FALSE;
            pthread_exit( NULL );
        }

    }

    g_bPicThreadRunning = FALSE;
    return NULL;
}


//////////////////////////////////////////////////////////////////////////////
// MphToKph - converts mph in kph expressed as tenths.
// KPH = (80*MPH)/5
//

int MphToKph( int nMph )
{
    return (nMph * 80)/5;
}

